// HndGoDrc Functions

//---------------------------------------------
// DRC Computation - Stack handling
//---------------------------------------------
int push(int c) {
	if (wmk < StackElements) {
		stack[++wmk] = c;
	} else {
		return -1;
	}			
}

int pop() {
	if (wmk > 0) {
		return stack[wmk--];
	} else {
		return -1;
	}
}

//----------------------------------------------------
// ScanForElement - Convert elements fron string to int
//----------------------------------------------------
int ScanForElement(string operand) {
int c;
	if (operand == "&&") {
		return LAND;
	} else {
		if (operand == "||") {
			return LOR;
		} else {
			if (operand == "!") {
				return NOT;
			} else {
				if (operand == "==") {
					return EQU;
				} else {
					if (operand == "<=") {
						return LTE;
					} else {
						if (operand == ">=") {
							return GTE;
						} else {
							if (operand == "!=") {
								return NTE;
							} else {
								if (operand == "<<") {
									return SHL;
								} else {
									if (operand == ">>") {
										return SHR;
									} else {
										if (operand == "~") {
											return INVERT;
										} else {
											if (operand == "&") {
												return BAND;
											} else {
												if (operand == "|") {
													return BOR;
												} else {
													if (operand == "^") {
														return BXOR;
													} else {	
														if (operand == "+") {
															return PLUS;
														} else {
															if (operand == "-") {
																return MINUS;
															} else {
																if (operand == "*") {
																	return MPLY;
																} else {
																	if (operand == "/") {
																		return DIVD;
																	} else {
																		if (operand == "%") {
																			return MODULO;
																		} else {
																			if (operand == "c") {
																				return -rpnC;
																			} else {
																				if (operand == "i") {
																					return -rpnI;
																				} else {
																					if (operand == "key") {
																						return -rpnKEY;
																					} else {
																						// No valid operand found. Maybe a number?
																						return (-1);
																					}
																				}	
																			}	
																		}	
																	}	
																}	
															}	
														}	
													}	
												}	
											}	
										}	
									}	
								}
							}	
						}
					}
				}	
			}	
		}	
	}
}


//---------------------------------------------
// ApplyRPNFormula - Applies RPN on C
//---------------------------------------------
int ApplyRPNFormula() {
int c, FirstOP, SecondOP, Result;
	for (c = 0; c < RPNwmk; c++) {
		// Let's see if we have an operator...
		if (RPNk[c] == OPERATOR) {		
			// Yes; this is an operator.
			// Let's pop the last two operators from the stack
			SecondOP = pop();
			FirstOP = pop();
			// Let's see which operator we have to apply...
			switch (RPN[c]) {
				// Unary operators
				case INVERT:
					// This is an unary operator, so we push the first operator back
					push (FirstOP);
					Result = ~SecondOP;
					break;

				case NOT:
					// This is an unary operator, so we push the first operator back
					push (FirstOP);
					if (SecondOP) {
						Result = False;
					} else {
						Result = True;
					}	
					break;
				
				// Binary operators
				case LAND:
					if ((FirstOP) && (SecondOP)) {
						Result = True;
					} else {
						Result = False;
					}		
					break;
				
				case LOR:
					if ((FirstOP) || (SecondOP)) {
						Result = True;
					} else {
						Result = False;
					}		
					break;
				
				case EQU:
					if (FirstOP == SecondOP) {
						Result = True;
					} else {
						Result = False;
					}		
					break;

				case LTE:
					if (FirstOP <= SecondOP) {
						Result = True;
					} else {
						Result = False;
					}		
					break;

				case GTE:
					if (FirstOP >= SecondOP) {
						Result = True;
					} else {
						Result = False;
					}		
					break;

				case NTE:
					if (FirstOP != SecondOP) {
						Result = True;
					} else {
						Result = False;
					}		
					break;

				case SHL:
					Result = (FirstOP << SecondOP);
					break;

				case SHR:
					Result = (FirstOP >> SecondOP);
					break;

				case BAND:
					Result = (FirstOP & SecondOP);
					break;

				case BOR:
					Result = (FirstOP | SecondOP);
					break;

				case BXOR:
					Result = (FirstOP ^ SecondOP);
					break;

				case PLUS:
					Result = (FirstOP + SecondOP);
					break;

				case MINUS:
					Result = (FirstOP - SecondOP);
					break;

				case MPLY:
					Result = (FirstOP * SecondOP);
					break;

				case DIVD:
					Result = (FirstOP / SecondOP);
					break;

				case MODULO:
					Result = (FirstOP % SecondOP);
					break;

				default:	
			}
			// Well, let's push the result onto the stack...
			push(Result);
		} else {
			// May be an operand or a variable, let's check...
			if (RPN[c] == -rpnC) {
				// C variable
				push(C);
			} else {
				if (RPN[c] == -rpnI) {
					// I variable
					push(I);
				} else {
					if (RPN[c] == -rpnKEY) {
						// KEY variable
						push(KEY);
					} else {
						// This is a number...	
						push(RPN[c]);
					}	
				}	
			}
		}
	}
	KEY = Result;
}


//---------------------------------------------
// DRC Computation - Formula parsing
//---------------------------------------------
int ParseRPN(string s) {
int i, c, integer, opTF, iOP;
string element = "";
	// We add an extra space, so the last operand/operator will be parsed without more effort...
	s = s + " ";
	opTF = False;
	for (c=0; c<strlen(s); c++) {
		if (s@[c] == " ") {
			// Space found; is there an element?
			if (opTF) {
				// Yes, we have an element. Let's check it...
				iOP = ScanForElement(element);
				if (iOP == -1) {
					// The supposed operator is unrecognized: maybe a numeric operand?
					for (i=0; i<strlen(element); i++) {
						if (strstr("-0123456789", element@[i], 0) == -1) {
							// It's not a number, we have an error
							return (False);
						}
					}
					// Yes, this is a numeric value; we have to store it (converting to Integer)
					iOP = element;
					RPN[RPNwmk] = iOP;
					// Tag it as an operand
					RPNk[RPNwmk] = OPERAND;
					RPNwmk++;
					element = "";
					opTF = False;
				} else {
					if (iOP < -1) {
						// If iOP less than -1, then ScanForElement returned a variable
						// We have to store as an operand
						RPN[RPNwmk] = iOP;
						// Tag it as an operand
						RPNk[RPNwmk] = OPERAND;
						RPNwmk++;
						element = "";
						opTF = False;
					} else {
						// Yes, it's a valid operator. Let's store it...
						RPN[RPNwmk] = iOP;
						// Tag it as an operator
						RPNk[RPNwmk] = OPERATOR;
						RPNwmk++;
						element = "";
						opTF = False;
					}	
				}
			} else {
				// No, we simply skip this space (initial? double in string? mah...)
			}
		} else {
			opTF = True;
			element = element + s@[c]; 
		}
	}
	return (True);
}


//---------------------------------------------
// DRC Computation - Returns the calculated DRC
//---------------------------------------------
string CalculateDRC(string lCDUI, string lFormula) {
int i, f, rc;
string sKEY;

	// Let's clear some variables...
	I = C = KEY = 0;

	// Let's allocate some memory to host CDUI's characters
	pCDUI = malloct(11, 'c');
	
	// First for all, check for CDUI's length & trim if necessary 
	// (only 10 digits are allowed)
	if (strlen(lCDUI) > 10) {
		// Too long, let's consider the first & the last 5 digits
		lCDUI = strleft(lCDUI, 5) + strright(lCDUI, 5);
	}
	// Let's set the CDUI Length
	CDUILth = strlen(lCDUI);
	
	// Let's separate each CDUI's character in the array pointed by pCDUI
	strtoc(lCDUI, pCDUI);

	// Let's parse RPN string (converted to lowercase)...
	if (ParseRPN(strlwr(lFormula)) == False) {
		// We have an error in the string...
		alert ("Alert: error in RPN string...");
		free (pCDUI);
		return "";
	} else {	
		// All seems correct. Let's start the calculation loop
		// for each character in CDUI using public I pointer
		for (I = 0; I < CDUILth; I++) {
			C = pCDUI[I];
			ApplyRPNFormula();	// !!! This function works on public variables...
		}
		
		// Let's check the value of KEY to correct if necessary
		// some code...
		free (pCDUI);
		while (KEY > 65535) {
			KEY = KEY - 65536;
		}
		while (KEY < 0) {
			KEY = KEY + 65536;
		}
		sKEY = KEY; // convert to string and adjust the length to five chars
		while (strlen(sKEY) < 5) {
			sKEY = "0" + sKEY;
		}
		return (sKEY); 
	}
}


